@@ -20,6 +20,37 @@ class ManagerUtilsQuerySet(QuerySet):
         """
         return {obj.id: obj for obj in self}
 
+    def bulk_upsert(self, objs, unique_fields, update_fields):
+        """
+        Performs a bulk update or insert on a queryset.
+        """
+        if not unique_fields:
+            raise ValueError('Must provide unique_fields argument')
+        if not update_fields:
+            raise ValueError('Must provide update_fields argument')
+
+        # Create a look up table for all of the objects in the queryset keyed on the unique_fields
+        model_obj_dict = {
+            tuple(getattr(model_obj, field) for field in unique_fields): model_obj for model_obj in self
+        }
+
+        # Find all of the objects to update and all of the objects to create
+        model_objs_to_update, model_objs_to_create = [], []
+        for obj in objs:
+            model_obj = model_obj_dict.get(tuple(obj[field] for field in unique_fields), None)
+            if model_obj is None:
+                # If the object needs to be created, make a new instance of it
+                model_objs_to_create.append(self.model(**obj))
+            else:
+                # If the object needs to be updated, update its fields
+                for field in update_fields:
+                    setattr(model_obj, field, obj[field])
+                model_objs_to_update.append(model_obj)
+
+        # Apply bulk updates and creates
+        self.model.objects.bulk_update(model_objs_to_update, update_fields)
+        self.model.objects.bulk_create(model_objs_to_create)
+
     def get_or_none(self, **query_params):
         """
         Get an object or return None if it doesn't exist.
@@ -65,7 +96,6 @@ class ManagerUtilsMixin(object):
             A dictionary of objects from the queryset or manager that is keyed on the objects' IDs.
 
         Examples:
-
             TestModel.objects.create(int_field=1)
             TestModel.objects.create(int_field=2)
 
@@ -74,6 +104,76 @@ class ManagerUtilsMixin(object):
         """
         return self.get_queryset().id_dict()
 
+    def bulk_upsert(self, objs, unique_fields, update_fields):
+        """
+        Performs a bulk update or insert on a list of dictionaries. Matches all objects in the queryset
+        with the objs provided using the field values in unique_fields.
+        If an existing object is matched, it is updated with the values from the provided objects. Objects
+        that don't match anything are bulk inserted.
+
+        Args:
+            objs: A list of dictionaries that have fields corresponding to the model in the manager.
+            unique_fields: A list of fields that are used to determine if an object in objs matches a model
+                from the queryset.
+            update_fields: A list of fields used from the objects in objs as fields when updating existing
+                models.
+
+        Signals: Emits a post_bulk_operation when a bulk_update or a bulk_create occurs.
+
+        Examples:
+            # Start off with no objects in the database. Call a bulk_upsert on the TestModel, which includes
+            # a char_field, int_field, and float_field
+            TestModel.objects.bulk_upsert([
+                {'float_field': 1.0, 'char_field': '1', 'int_field': 1},
+                {'float_field': 2.0, 'char_field': '2', 'int_field': 2},
+                {'float_field': 3.0, 'char_field': '3', 'int_field': 3},
+            ], ['int_field'], ['char_field'])
+
+            # All objects should have been created
+            print TestModel.objects.count()
+            3
+
+            # Now perform a bulk upsert on all the char_field values. Since the objects existed previously
+            # (known by the int_field uniqueness constraint), the char fields should be updated
+            TestModel.objects.bulk_upsert([
+                {'float_field': 1.0, 'char_field': '0', 'int_field': 1},
+                {'float_field': 2.0, 'char_field': '0', 'int_field': 2},
+                {'float_field': 3.0, 'char_field': '0', 'int_field': 3},
+            ], ['int_field'], ['char_field'])
+
+            # No more new objects should have been created, and every char field should be 0
+            print TestModel.objects.count(), TestModel.objects.filter(char_field='-1').count()
+            3, 3
+
+            # Do the exact same operation, but this time add an additional object that is not already
+            # stored. It will be inserted.
+            TestModel.objects.bulk_upsert([
+                {'float_field': 1.0, 'char_field': '1', 'int_field': 1},
+                {'float_field': 2.0, 'char_field': '2', 'int_field': 2},
+                {'float_field': 3.0, 'char_field': '3', 'int_field': 3},
+                {'float_field': 4.0, 'char_field': '4', 'int_field': 4},
+            ], ['int_field'], ['char_field'])
+
+            # There should be one more object
+            print TestModel.objects.count()
+            4
+
+            # Note that one can also do the upsert on a queryset. Perform the same data upsert on a
+            # filter for int_field=1. In this case, only one object has the ability to be updated.
+            # All of the other objects will be inserted
+            TestModel.objects.filter(int_field=1).bulk_upsert([
+                {'float_field': 1.0, 'char_field': '1', 'int_field': 1},
+                {'float_field': 2.0, 'char_field': '2', 'int_field': 2},
+                {'float_field': 3.0, 'char_field': '3', 'int_field': 3},
+                {'float_field': 4.0, 'char_field': '4', 'int_field': 4},
+            ], ['int_field'], ['char_field'])
+
+            # There should be three more objects
+            print TestModel.objects.count()
+            7
+        """
+        return self.get_queryset().bulk_upsert(objs, unique_fields, update_fields)
+
     def bulk_create(self, objs, batch_size=None):
         """
         Overrides Django's bulk_create function to emit a post_bulk_operation signal when bulk_create
@@ -139,39 +239,39 @@ class ManagerUtilsMixin(object):
             updates: These values are updated when the object is updated. They also override any
                 values provided in the defaults when inserting the object.
             **kwargs: These values provide the arguments used when checking for the existence of
-                the object.  They are used in a similar manner to Django's get_or_create function.
+                the object. They are used in a similar manner to Django's get_or_create function.
 
         Returns: A tuple of the upserted object and a Boolean that is True if it was created (False otherwise)
 
         Examples:
-        # Upsert a test model with an int value of 1. Use default values that will be given to it when created
-        model_obj, created = TestModel.objects.upsert(int_field=1, defaults={'float_field': 2.0})
-        print created
-        True
-        print model_obj.int_field, model_obj.float_field
-        1, 2.0
+            # Upsert a test model with an int value of 1. Use default values that will be given to it when created
+            model_obj, created = TestModel.objects.upsert(int_field=1, defaults={'float_field': 2.0})
+            print created
+            True
+            print model_obj.int_field, model_obj.float_field
+            1, 2.0
 
-        # Do an upsert on that same model with different default fields. Since it already exists, the defaults
-        # are not used
-        model_obj, created = TestModel.objects.upsert(int_field=1, defaults={'float_field': 3.0})
-        print created
-        False
-        print model_obj.int_field, model_obj.float_field
-        1, 2.0
+            # Do an upsert on that same model with different default fields. Since it already exists, the defaults
+            # are not used
+            model_obj, created = TestModel.objects.upsert(int_field=1, defaults={'float_field': 3.0})
+            print created
+            False
+            print model_obj.int_field, model_obj.float_field
+            1, 2.0
 
-        # In order to update the float field in an existing object, use the updates dictionary
-        model_obj, created = TestModel.objects.upsert(int_field=1, updates={'float_field': 3.0})
-        print created
-        False
-        print model_obj.int_field, model_obj.float_field
-        1, 3.0
+            # In order to update the float field in an existing object, use the updates dictionary
+            model_obj, created = TestModel.objects.upsert(int_field=1, updates={'float_field': 3.0})
+            print created
+            False
+            print model_obj.int_field, model_obj.float_field
+            1, 3.0
 
-        # You can use updates on a newly created object that will also be used as initial values.
-        model_obj, created = TestModel.objects.upsert(int_field=2, updates={'float_field': 4.0})
-        print created
-        True
-        print model_obj.int_field, model_obj.float_field
-        2, 4.0
+            # You can use updates on a newly created object that will also be used as initial values.
+            model_obj, created = TestModel.objects.upsert(int_field=2, updates={'float_field': 4.0})
+            print created
+            True
+            print model_obj.int_field, model_obj.float_field
+            2, 4.0
         """
         obj, created = self.model.objects.get_or_create(defaults=defaults or {}, **kwargs)
 
